import argparse
import json
import math
import time
from tqdm import tqdm
import torch
from torch.utils.data import DataLoader, RandomSampler
from torch.utils.data.distributed import DistributedSampler
import deepspeed
import logging
from utils import print_trainable_parameters, print_rank_0, to_device, set_random_seed, save_model
from utils import DataCollator
from peft import LoraConfig, get_peft_model
from model import MODE

# 初始化日志记录
logging.basicConfig(
    filename="training.log",  # 设置日志文件路径
    filemode="w",  # 写入模式
    level=logging.INFO,  # 设置日志记录级别
    format="%(asctime)s - %(levelname)s - %(message)s"  # 设置日志格式
)
logger = logging.getLogger(__name__)

# 清理未使用的显存
torch.cuda.empty_cache()


def parse_args():
    parser = argparse.ArgumentParser()
    # 可修改的路径和参数
    parser.add_argument("--model_name_or_path", type=str, default="/root/autodl-tmp/ChatGLM-Finetuning-master/chatglm-6b",help="Path to the pretrained model")
    parser.add_argument("--tokenizer_path", type=str, default="/root/autodl-tmp/ChatGLM-Finetuning-master/chatglm-6b",help="Path to the tokenizer")
    parser.add_argument("--train_path", type=str, default="data/spo_0.json", help="Path to the training data")
    parser.add_argument("--output_dir", type=str, default="./output-glm", help="Output directory for saving models")
    
    # 其他训练参数
    parser.add_argument("--max_len", type=int, default=1024, help="")
    parser.add_argument("--max_src_len", type=int, default=256, help="")
    parser.add_argument("--per_device_train_batch_size", type=int, default=4, help="")
    parser.add_argument("--learning_rate", type=float, default=1e-3, help="")
    parser.add_argument("--weight_decay", type=float, default=0.1, help="")
    parser.add_argument("--num_train_epochs", type=int, default=1, help="")
    parser.add_argument("--gradient_accumulation_steps", type=int, default=1, help="")
    parser.add_argument("--warmup_ratio", type=float, default=0.1, help="")
    parser.add_argument("--mode", type=str, default="glm", help="Model type: glm, glm2, glm3")
    parser.add_argument("--train_type", type=str, default="lora", help="Training type: lora, freeze, ptuning, all")
    parser.add_argument("--seed", type=int, default=1234, help="Random seed")
    parser.add_argument("--gradient_checkpointing", action='store_true', help="Enable gradient checkpointing")
    parser.add_argument("--ds_file", type=str, default="ds_zero3_no_offload.json", help="DeepSpeed config file")
    parser.add_argument("--show_loss_step", default=10, type=int, help="Show loss every n steps")
    parser.add_argument("--save_model_step", default=None, type=int, help="Save model every n steps")
    parser.add_argument("--local_rank", type=int, default=-1, help="Local rank for distributed training")
    parser = deepspeed.add_config_arguments(parser)
    return parser.parse_args()


def main():
    args = parse_args()

    device = torch.device(f"cuda:{args.local_rank}" if args.local_rank != -1 else "cuda" if torch.cuda.is_available() else "cpu")
    set_random_seed(args.seed)

    # DeepSpeed 配置
    with open(args.ds_file, "r", encoding="utf-8") as fh:
        ds_config = json.load(fh)

    # 动态调整 DeepSpeed 配置
    ds_config['train_micro_batch_size_per_gpu'] = args.per_device_train_batch_size
    ds_config['gradient_accumulation_steps'] = args.gradient_accumulation_steps
    ds_config['train_batch_size'] = args.per_device_train_batch_size * args.gradient_accumulation_steps

    # 加载分词器
    tokenizer = MODE[args.mode]["tokenizer"].from_pretrained(args.tokenizer_path)
    logger.info(f"Loaded tokenizer from {args.tokenizer_path}")

    # 加载模型
    model = MODE[args.mode]["model"].from_pretrained(args.model_name_or_path)
    logger.info(f"Loaded model from {args.model_name_or_path}")
    
    # 启用梯度检查点（如果启用--gradient_checkpointing参数）
    if args.gradient_checkpointing:
        print_rank_0("Enabling gradient checkpointing")
        model.gradient_checkpointing_enable()

    # 数据加载
    train_dataset = MODE[args.mode]["dataset"](args.train_path, tokenizer, args.max_len, args.max_src_len, False)

    # 分布式训练时使用 DistributedSampler
    if args.local_rank != -1:
        train_sampler = DistributedSampler(train_dataset, num_replicas=torch.cuda.device_count(), rank=args.local_rank)
    else:
        train_sampler = RandomSampler(train_dataset)

    data_collator = DataCollator(tokenizer)
    train_dataloader = DataLoader(train_dataset, collate_fn=data_collator, sampler=train_sampler,
                                  batch_size=args.per_device_train_batch_size)
    logger.info(f"Training dataset size: {len(train_dataset)}")

    # 优化器和调度器配置
    ds_config["optimizer"]["params"]["lr"] = args.learning_rate
    ds_config["scheduler"]["params"]["warmup_num_steps"] = int(args.warmup_ratio * len(train_dataloader))

    # 初始化 DeepSpeed
    model, optimizer, _, lr_scheduler = deepspeed.initialize(model=model, args=args, config=ds_config)

    # 开始训练循环
    model.train()
    start_time = time.time()
    global_step = 0
    total_steps = len(train_dataloader) * args.num_train_epochs

    for epoch in range(args.num_train_epochs):
        logger.info(f"Starting epoch {epoch + 1}/{args.num_train_epochs}")
        for step, batch in tqdm(enumerate(train_dataloader), total=len(train_dataloader)):
            batch = to_device(batch, device)

            step_start = time.time()
            outputs = model(**batch)
            loss = outputs.loss
            model.backward(loss)
            model.step()
            step_time = time.time() - step_start

            global_step += 1
            logger.info(f"Step {global_step}/{total_steps}, Loss={loss.item()}, Step Time={step_time:.2f}s")

    # 训练完成
    total_time = time.time() - start_time
    logger.info(f"Training completed in {total_time:.2f} seconds")
    save_model(model, tokenizer, args.output_dir)


if __name__ == "__main__":
    main()
